use std::{cell::RefCell, fmt::Display, rc::Rc};

use ordered_float::OrderedFloat;

use crate::{callable::Callable, instance, token::Literal};

/// Represents the type of a callable value in Lox.
#[derive(Clone, Debug, PartialEq)]
pub enum CallableType {
    /// A regular function or method
    Function,
    /// A class (which is callable to create instances)
    Class(crate::class::Class),
}

/// Concrete runtime value in the Lox interpreter.
#[derive(Clone, Debug, PartialEq)]
pub enum Value {
    /// Special value for function returns
    Return(Box<Value>),
    /// Function, method, or class
    Callable((Rc<dyn Callable>, CallableType)),
    /// String value
    String(String),
    /// Numeric value
    Number(OrderedFloat<f64>),
    /// Boolean value
    Boolean(bool),
    /// Class instance
    Instance(Rc<RefCell<instance::Instance>>),
    /// Nil value (null)
    Nil,
}

impl Display for Value {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            Value::Return(x) => write!(f, "{x}"),
            Value::Callable(x) => write!(f, "{}", x.0),
            Value::String(x) => write!(f, "{x}"),
            Value::Number(x) => write!(f, "{x}"),
            Value::Boolean(x) => write!(f, "{x}"),
            Value::Instance(x) => write!(f, "{}", x.borrow()),
            Value::Nil => write!(f, "nil"),
        }
    }
}

impl Value {
    /// Determine if a value is "truthy" according to Lox rules.
    /// Return false for nil or a false boolean, true otherwise.
    pub fn is_truthy(&self) -> bool {
        match self {
            Value::Nil => false,
            Value::Boolean(b) => *b,
            _ => true,
        }
    }
}

impl From<Literal> for Value {
    fn from(value: Literal) -> Self {
        match value {
            Literal::String(x) => Value::String(x),
            Literal::Number(x) => Value::Number(x),
            Literal::Boolean(x) => Value::Boolean(x),
            Literal::Nil => Value::Nil,
        }
    }
}
